.include "m163def.inc"

.equ MEM_FREELISTL  = 0x0060
.equ MEM_FREELISTH  = 0x0061    ; memory allocator free list
.equ SCH_PROCLISTL  = 0x0062
.equ SCH_PROCLISTH  = 0x0063    ; round-robin list of processes
.equ SCH_NPROC      = 0x0064    ; number of active processes in list
.equ UART_TX_PROCL  = 0x0065
.equ UART_TX_PROCH  = 0x0066    ; process using the UART transmitter
.equ UART_RX_PROCL  = 0x0067
.equ UART_RX_PROCH  = 0x0068    ; process using the UART receiver
.equ MUTEX          = 0x0069
.equ MUTEX_MEM      = 0         ; memory free list bit
.equ MUTEX_IPTX     = 1         ; SLIP packet transfer bit
.equ INT_TMP        = 0x006a    ; 5 byte temp space for interrupt handlers
.equ UPTIME         = 0x006f    ; 5 byte nr of ticks since boot
.equ FREE_RAM_START = 0x0074    ; first free RAM address
.equ FREE_RAM_END   = RAMEND+1  ; last free RAM address + 1
.equ INITSP         = RAMEND-2

.equ EE_IPADR       = EEPROMEND-3   ; 4-byte IP address of host
.equ EE_BAUDRATE    = EEPROMEND-4   ; baud rate divisor for UART
.equ EE_FREE_END    = EEPROMEND-4   ; last free EEPROM address + 1

.equ B2400          = 191
.equ B4800          = 95
.equ B9600          = 47
.equ B14400         = 31
.equ B19200         = 23
.equ B28800         = 15
.equ B38400         = 11
.equ B57600         = 7
.equ B76800         = 5
.equ B115200        = 3

.equ UART_BAUD      = B38400

.eseg
.org EE_IPADR
.db 10,0,0,1
.org EE_BAUDRATE
.db UART_BAUD

.equ LED_PORT       = PORTC
.equ LED_DDR        = DDRC


; MUTEX_WAIT rXX,b
; Wait for mutex bit b to be clear, then set it.
; rXX will be destroyed. XX must be 16 or more.
.macro MUTEX_WAIT
    in @0,SREG
    push @0
    cli
    lds @0,MUTEX
    bst @0,@1
    brtc PC+6
    pop @0
    out SREG,@0
    call sch_yield
    rjmp PC-11
    ori @0,1<<@1
    sts MUTEX,@0
    pop @0
    out SREG,@0
.endmacro

; MUTEX_LEAVE rXX,b
; Clear mutex bit b.
; rXX will be set to MUTEX, XX >= 16
.macro MUTEX_LEAVE
    in @0,SREG
    push @0
    cli
    lds @0,MUTEX
    andi @0,~(1<<@1)
    sts MUTEX,@0
    pop @0
    out SREG,@0
.endmacro

; SCH_INCPROC rXX
; Increase SCH_NPROC by one if TF == 1
; rXX destroyed
.macro SCH_INCPROC
    brtc PC+6
    lds @0,SCH_NPROC
    inc @0
    sts SCH_NPROC,@0
.endmacro

; SCH_WAKEUP
; Wake up the process in Y. If it's not already awake, also increase
; SCH_NPROC by 1.
; rXX destroyed, XX >= 16
.macro SCH_WAKEUP
    ldd @0,y+PROC_FLAGS
    bst @0,7
    andi @0,0x7f
    std y+PROC_FLAGS,@0
    brtc PC+6
    lds @0,SCH_NPROC
    inc @0
    sts SCH_NPROC,@0
.endmacro

; SCH_SAVEFAST
; Save X, Z and SREG to the first 5 bytes of INT_TMP
; 11 cycles
.macro SCH_SAVEFAST
    sts INT_TMP+0,xl
    sts INT_TMP+1,xh
    sts INT_TMP+2,zl
    sts INT_TMP+3,zh
    in xl,SREG
    sts INT_TMP+4,xl
.endmacro

; SCH_LOADFAST
; Load the info saved by SCH_SAVEFAST
; 11 cycles
.macro SCH_LOADFAST
    lds xl,INT_TMP+4
    out SREG,xl
    lds xl,INT_TMP+0
    lds xh,INT_TMP+1
    lds zl,INT_TMP+2
    lds zh,INT_TMP+3
.endmacro

; EEPROM_READ address,rXX
; Read a byte from the EEPROM address (constant) to rXX
; XX >= 16
.macro EEPROM_READ_IMM
    ldi @1,high(@0)
    out EEARH,@1
    ldi @1,low(@0)
    out EEARL,@1
    sbic EECR,EEWE
    rjmp PC-1
    sbi EECR,EERE
    in @1,EEDR
.endmacro

.cseg
.org 0
    jmp reset
    jmp int_ext0
    jmp int_ext1
    jmp int_tim2_comp
    jmp int_tim2_ovf
    jmp int_tim1_capt
    jmp int_tim1_compa
    jmp int_tim1_compb
    jmp int_tim1_ovf
    jmp int_tim0_ovf
    jmp int_spi_stc
    jmp int_uart_rxc
    jmp int_uart_dre
    jmp int_uart_txc
    jmp int_adc
    jmp int_ee_rdy
    jmp int_ana_comp
    jmp int_twi

int_ext0:
int_ext1:
int_tim2_comp:
int_tim2_ovf:
int_tim1_capt:
int_tim1_compa:
int_tim1_compb:
int_tim1_ovf:
;int_tim0_ovf:
int_spi_stc:
;int_uart_rxc:
;int_uart_dre:
int_uart_txc:
int_adc:
int_ee_rdy:
int_ana_comp:
int_twi:
    reti

reset:
    ldi r16,high(INITSP)        ; initialize SP to a memory area at the
    out SPH,r16                 ; top of the first memory block, which
    ldi r16,low(INITSP)         ; should remain unallocated until we can
    out SPL,r16                 ; create a proper process with sch_create

    ser r16
    out LED_PORT,r16
    out LED_DDR,r16             ; initialize and turn off LEDs

    call uart_init
    call mem_init
    call sch_init

    ldi r16,(1<<WDE)
    out WDTCR,r16               ; enable WDT, ~15ms

    sei

    ldi r16,0x16
    clr r17
    ldi r18,low(init)
    ldi r19,high(init)
    call sch_create

    jmp sch_yield

init:
    ldi r16,45
    mov r0,r16
    clr r1
    clr r2
    clr r3
    clr r4                      ; 45 ticks = 100 ms
    call sch_delay              ; wait for RS232 circuit to stabilize

init_receive:
    call ip_rx
    brne init_receive           ; error receiving? then ignore
    movw xl,zl                  ; X = pointer to packet buffer
    ldi r16,0x16                ; 22 byte data/stack space
    clr r17
    ldi r18,low(handle_packet)
    ldi r19,high(handle_packet)
    call sch_create             ; create a process at handle_packet
    rjmp init_receive           ; and wait for another packet

handle_packet:                  ; X points to packet buffer
    movw zl,xl
    ldd r16,z+IP_PROTO
    cpi r16,PROTO_ICMP
    breq handle_packet_icmp
    cpi r16,PROTO_UDP
    breq handle_packet_udp
handle_packet_drop:
    mov zl,xl
    call mem_free
    jmp sch_die

handle_packet_icmp:
    ldd r16,z+20+ICMP_TYPE
    cpi r16,8                   ; ICMP Echo request?
    breq handle_packet_echo
    rjmp handle_packet_drop     ; no, drop packet

handle_packet_udp:
    ldd r16,z+20+UDP_DSTPORT+0
    cpi r16,0
    brne handle_packet_udp0
    ldd r16,z+20+UDP_DSTPORT+1
    cpi r16,7
    breq handle_packet_udpecho
handle_packet_udp0:
    rjmp handle_packet_drop

handle_packet_echo:
    call ip_swap_adr            ; swap src and dest fields
    call ip_create_header       ; create IP header for reply
    clr r16
    std z+20+ICMP_TYPE,r16      ; type = 0 (Echo Reply)
    std z+20+ICMP_CODE,r16      ; code = 0
    std z+20+ICMP_CKSUM+0,r16
    std z+20+ICMP_CKSUM+1,r16   ; checksum = 0 for now
    ldd r17,z+IP_LEN+0
    ldd r16,z+IP_LEN+1
    subi r16,20
    sbci r17,0                  ; r17:16 = length of ICMP block
    adiw zl,20                  ; Z = start of ICMP header
    clr r0
    clr r1
    call ip_ocsum               ; compute ICMP checksum
    movw zl,xl
    com r0
    com r1                      ; r1:r0 = one's complement of sum
    std z+20+ICMP_CKSUM+0,r1
    std z+20+ICMP_CKSUM+1,r0    ; store computed checksum
    call ip_tx                  ; transmit packet
    rjmp handle_packet_drop

handle_packet_udpecho:
    call ip_swap_adr            ; swap src and dest fields
    call ip_create_header       ; create IP header for reply
    clr r16
    std z+20+UDP_CKSUM+0,r16
    std z+20+UDP_CKSUM+1,r16    ; TODO: implement checksum?
    ldd r16,z+20+UDP_SRCPORT+0
    std z+20+UDP_DSTPORT+0,r16
    ldd r16,z+20+UDP_SRCPORT+1
    std z+20+UDP_DSTPORT+1,r16  ; reply to source port
    clr r16
    std z+20+UDP_SRCPORT+0,r16
    ldi r16,7
    std z+20+UDP_SRCPORT+1,r16  ; new source port = 7
    call ip_tx                  ; transmit packet
    rjmp handle_packet_drop

.include "mem.asm"
.include "sch.asm"
.include "uart.asm"
.include "ip.asm"

